
import datetime
import traceback
from typing import Tuple, List, Optional

import numpy as np
import pandas as pd
import statsmodels.api as sm
from pandas.tseries.offsets import DateOffset

from .derived import DerivedFactor
from .basic_factors import BasicFactor
from .utils import TradingDayList, StockIndustryInfoFactor, FundHoldStockFactor
from .factor_types import MixFactorBase


class StockFactorApi:

    _DECOMP_RET_DATE_OFFSET = DateOffset(years=1)
    _WINDOW_LEN = int(60 / 5)

    @staticmethod
    def get_stock_factor_ret(factor_names: Tuple[str], universe: str) -> Optional[pd.DataFrame]:
        rets: List[pd.Series] = []
        for one in factor_names:
            try:
                factor = DerivedFactor._registry[one]
            except KeyError as e:
                print(f'[stock_factor_api] invalid stock factor name {one} (err_msg){e} (universe){universe}')
                continue
            f_ret = factor.get_ret(universe=universe)
            if f_ret is not None:
                rets.append(f_ret)
            else:
                print(f'[stock_factor_api] retrieve return of {one} failed (universe){universe}')
        if rets:
            return pd.concat(rets, axis=1)

    @staticmethod
    def decomp_ret(fund_ret: pd.Series, factor_names: Tuple[str], universe: str) -> Optional[Tuple[pd.Series, pd.DataFrame]]:

        def _lambda_rolling_reg(x, whole, fund_name):
            # x是基金每日收益率, 对相应日期的因子收益做回归(加前置的常数项)(使用前WINDOW_LEN日)
            results = sm.OLS(x[:-1], sm.add_constant(whole.drop(columns=fund_name).loc[x.index[:-1], :])).fit()
            # 回归结果中包含常数项系数，这里只取另外几个indicator的系数(即上边的因子收益)
            # T+1日各因子收益的贡献 = T+1日因子收益 * 前WINDOW_LEN日回归系数
            contribution = whole.drop(columns=fund_name).loc[x.index[-1], :] * results.params[1:]
            # T+1日特异收益率 = T+1日基金收益 - T+1日各因子收益的贡献和
            spec_ret = x.array[-1] - sum(contribution)

            # 记录T+1日各因子收益的贡献, 特异收益率, 基金收益率等信息
            fund_attribution_list.append(contribution.append(pd.Series({'spec': spec_ret, 'datetime': x.index.array[-1]})))
            return 0

        ret_rg = StockFactorApi.get_stock_factor_ret(factor_names, universe)
        if ret_rg is None:
            print(f'[stock_factor_api] retrieve return of {factor_names} failed')
            return
        try:
            fund_ret = fund_ret.dropna()
            # 是否只使用部分数据计算因子暴露
            # cal_start_date = (fund_ret.index.array[-1] - StockFactorApi._DECOMP_RET_DATE_OFFSET).date()
            # if fund_ret.index.array[0] > cal_start_date:
            #     print(f'[stock_factor_api] not enough data of fund {fund_ret.name} (start_date){fund_ret.index.array[0]}')
            #     return
            # filtered_fund_ret = fund_ret[fund_ret.index >= cal_start_date]
            # if filtered_fund_ret.empty:
            #     print(f'[stock_factor_api] no data of fund {fund_ret.name} to decomp (start_date){fund_ret.index.array[0]} (end_date){fund_ret.index.array[-1]}')
            #     return
            # rg = filtered_fund_ret.to_frame().join(ret_rg).ffill().dropna()
            # 计算因子收益曲线(包含特异收益率)
            rg = fund_ret.to_frame().join(ret_rg).ffill()
            rg = rg.set_axis(pd.to_datetime(rg.index, infer_datetime_format=True)).resample('1W').sum(min_count=3)
            rg = rg.set_axis(rg.index.date)
            rg = rg[rg.notna().all(axis=1)]
            fund_attribution_list = []
            # 以下为手写的rolling, 但实际和直接用rolling的时间差不多
            # for i in range(rg.shape[0]):
            #     end_loc = i + StockFactorApi._WINDOW_LEN
            #     if end_loc + 1 >= rg.shape[0]:
            #         break
            #     data_in_window = rg.iloc[i:end_loc, :]
            #     # x是基金每日收益率, 对相应日期的因子收益做回归(加前置的常数项)(使用前WINDOW_LEN日)
            #     results = sm.OLS(data_in_window[fund_ret.name], sm.add_constant(data_in_window.drop(columns=fund_ret.name))).fit()
            #     # 回归结果中包含常数项系数，这里只取另外几个indicator的系数(即上边的因子收益)
            #     # T+1日各因子收益的贡献 = T+1日因子收益 * 前WINDOW_LEN日回归系数
            #     contribution = rg.iloc[end_loc, :].drop(fund_ret.name) * results.params[1:]
            #     # T+1日特异收益率 = T+1日基金收益 - T+1日各因子收益的贡献和
            #     spec_ret = rg.iloc[end_loc, :].at[fund_ret.name] - sum(contribution)

            #     # 记录T+1日各因子收益的贡献, 特异收益率, 基金收益率等信息
            #     fund_attribution_list.append(contribution.append(pd.Series({'spec': spec_ret, 'datetime': rg.index.array[end_loc]})))
            rg[fund_ret.name].rolling(window=StockFactorApi._WINDOW_LEN+1).apply(_lambda_rolling_reg, kwargs={'whole': rg, 'fund_name': fund_ret.name})
            factor_ret_decomp = pd.DataFrame(fund_attribution_list).set_index('datetime').cumsum()
            # factor_ret_decomp = (np.exp(factor_ret_decomp.drop(columns='spec_ret')) - 1).join(factor_ret_decomp.spec_ret)
            # 计算基金全时长的因子暴露
            Y = rg.iloc[:, 0]
            X = sm.add_constant(rg.iloc[:, 1:])
            model1 = sm.OLS(Y, X)
            resu1 = model1.fit()
            return pd.Series(resu1.params[1:], index=ret_rg.columns, name=fund_ret.name), factor_ret_decomp
        except Exception as e:
            print(f'[stock_factor_api] abnormal error when decomp return (err_msg){e}')
            traceback.print_exc()
            return

    @staticmethod
    def fac_pos(fund_id: str, factor_names: Tuple[str], report_date: datetime.date) -> Optional[pd.DataFrame]:
        df = FundHoldStockFactor().get()
        if df is None:
            print(f'[fac_pos] failed to get data of fund hold stock')
            return
        df = df[(df.fund_id == fund_id) & (df.datetime == report_date)]
        df = df[df.notna().all(axis=1)]
        if df.empty:
            print(f'[fac_pos] empty df, can not calc (fund_id){fund_id} (report_date){report_date}')
            return

        s_list: List[pd.Series] = []
        try:
            for one in factor_names:
                try:
                    factor = DerivedFactor._registry[one]
                except KeyError as e:
                    print(f'[fac_pos] invalid stock factor name {one} (err_msg){e}')
                    continue
                f_val = factor.get()
                if f_val is None:
                    print(f'[fac_pos] retrieve value of {one} failed')
                    continue
                anjou = []
                weight = []
                for i in range(1, 11):
                    lis = df['rank'+str(i)+'_stock_code'].values
                    mult1 = pd.DataFrame(f_val[lis].iloc[-1])
                    mult1.columns = ['rank'+str(i)+'_fac']
                    mult2 = pd.DataFrame(df['rank'+str(i)+'_stockweight']).set_index(df['rank'+str(i)+'_stock_code'])
                    mult2.columns = ['rank'+str(i)+'_fac']
                    w = mult2.copy()
                    w.index = df.fund_id.to_list()
                    weight.append(w)
                    louis = (mult1 * mult2).replace({np.nan: 0})
                    louis.index = df.fund_id.to_list()
                    anjou.append(louis)
                res = (pd.concat(anjou, axis=1).sum(axis=1) / pd.concat(weight, axis=1).sum(axis=1)).rename(one)
                s_list.append(res)
        except Exception as e:
            print(f'[fac_pos] abnormal error when calc fac pos (err_msg){e}')
            traceback.print_exc()
            return
        res_df = pd.DataFrame(s_list)
        if res_df.empty:
            return
        return res_df[fund_id]


class StockFactorUpdater:

    @staticmethod
    def do_update_everyday(end_date: str, universe_list: Optional[Tuple[str]]) -> List[str]:
        def do_one_factor_update(obj) -> bool:
            if obj.name in factor_done:
                return True
            # 遍历每一个依赖, 先进行所依赖因子的更新, 这里递归可能很深
            ret = True
            for dep in obj._deps:
                ret &= do_one_factor_update(dep)
            if not isinstance(obj, (BasicFactor, DerivedFactor)):
                print(f'invalid factor {obj.name} (type){type(obj)}')
                nonlocal failed_f
                failed_f.append(obj.name)
                return False
            if isinstance(obj, MixFactorBase):
                if universe_list is not None:
                    for universe in universe_list:
                        obj.calc(universe)
                        ret &= obj.save(universe)
                        if ret:
                            print(f'save {obj.name} on {universe} done')
                        else:
                            print(f'save {obj.name} on {universe} failed')
            else:
                obj.calc()
                ret &= obj.save()
            if ret:
                print(f'save {obj.name} done')
            else:
                print(f'save {obj.name} failed')
            if isinstance(obj, DerivedFactor) and universe_list is not None:
                for universe in universe_list:
                    obj.calc_ret(universe)
                    ret &= obj.save_ret(universe)
                    if ret:
                        print(f'save ret of {obj.name} on {universe} done')
                    else:
                        print(f'save ret of {obj.name} on {universe} failed')
            factor_done.add(obj.name)
            return ret

        # 从derived因子驱动, 更新每一个因子
        failed_f: List[str] = []
        print('to update all factors')
        for one in (TradingDayList, StockIndustryInfoFactor, FundHoldStockFactor):
            obj = one()
            try:
                obj.calc()
                if isinstance(obj, TradingDayList):
                    # 不存trading day list, 但需要trim一下
                    obj.trim_by_date(end_date)
                elif not isinstance(obj, StockIndustryInfoFactor):
                    obj.save()
            except Exception as e:
                print(f'update util factor {obj.name} failed (err_msg){e}')
                traceback.print_exc()
                failed_f.append(obj.name)
        print('update utils done')
        factor_done = set()
        for obj in DerivedFactor._registry.values():
            try:
                result = do_one_factor_update(obj)
            except Exception as e:
                print(f'update derived factor {obj.name} failed (err_msg){e}')
                traceback.print_exc()
                failed_f.append(obj.name)
            else:
                if not result:
                    failed_f.append(obj.name)
        print('update all factors done')
        return failed_f



import pandas as pd
import traceback
from sqlalchemy import distinct, func
from ...wrapper.mysql import ViewDatabaseConnector
from ...view.view_models import ManagerInfoCollection


class ManagerInfoCollectionProcessor:

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def collection_daily(self):
        mng_info = [
            {
                'mng_id':'16E872',
                'mng_name':'萧楠',
                'introduction':'消费之王，坚守价值投资，屡获奖项',
                'fund_type': 'stock',
            },
            {
                'mng_id':'170F31',
                'mng_name':'谢治宇',
                'introduction':'自下而上精选个股，行业均衡，成长投资，代表作7年6倍',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7CABBB',
                'mng_name':'袁芳',
                'introduction':'擅长行业消费+科技，风格均衡，高换手分散持股',
                'fund_type': 'stock',
            },
            {
                'mng_id':'17109A',
                'mng_name':'曲扬',
                'introduction':'投资风格以稳健著称，获得第七届英华奖五年期股票投资最佳基金经理奖项',
                'fund_type': 'stock',
            },
            {
                'mng_id':'16E885',
                'mng_name':'张坤',
                'introduction':'收益优秀，8年6倍，700亿规模',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7C3A95',
                'mng_name':'杨浩',
                'introduction':'TMT行业有着深入的研究，交银三剑客',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7B55CF',
                'mng_name':'谭小兵',
                'introduction':'任职四年半回报230%',
                'fund_type': 'stock',
            },
            {
                'mng_id':'18377A',
                'mng_name':'王君正',
                'introduction':'通过稳健的投资框架，获得过金牛奖、明星基金奖和金基金奖共6座，业内少有的大满贯基金经理',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7BE82B',
                'mng_name':'何帅',
                'introduction':'绝对收益；成长价值；性价比为王；长期优异；交银三剑客',
                'fund_type': 'stock',
            },
            {
                'mng_id':'78F3D1',
                'mng_name':'孙伟',
                'introduction':'连续5年金牛奖，全市场获此荣誉的基金经理仅3人三次获得明星基金奖',
                'fund_type': 'stock',
            },
            {
                'mng_id':'19A22E',
                'mng_name':'张清华',
                'introduction':'股债双修研究深入，择时能力突出',
                'fund_type': 'bond',
            },
            {
                'mng_id':'01109',
                'mng_name':'桂跃强',
                'introduction':'股债配置能力强，投资权益的风格就属于稳健型，擅长使用权益类资产进行稳定增强',
                'fund_type': 'bond',
            },
            {
                'mng_id':'4F9879',
                'mng_name':'姚秋',
                'introduction':'股债全能的投资界“名将”，宏观层面把握机会，追求长期稳健回报资产配置能力出色',
                'fund_type': 'bond',
            },
            {
                'mng_id':'4F9865',
                'mng_name':'张雅君',
                'introduction':'金牛债券基金经理，同类产品连续六年正收益，金牛基金、明星基金，五星债券基金',
                'fund_type': 'bond',
            },
            {
                'mng_id':'C3B3E7',
                'mng_name':'蒋利娟',
                'introduction':'第六届中国基金业英华奖三年期纯债投资最佳基金经理、泰康资产公募事业部固收投资负责人',
                'fund_type': 'bond',
            },
            {
                'mng_id':'00707',
                'mng_name':'苏玉平',
                'introduction':'五年期二级债最佳基金经理奖”，连续三年在英华奖评选中榜上有名',
                'fund_type': 'bond',
            },
            {
                'mng_id':'17036A',
                'mng_name':'于泽雨',
                'introduction':'2019年金牛奖最佳人气金牛基金经理得主、擅长分析基本面 擅长长周期操作',
                'fund_type': 'bond',
            },
            {
                'mng_id':'7A8375',
                'mng_name':'张雪',
                'introduction':'擅长使用久期择时，杠杆择时，注重风险调整后收益，夏普比率2.58',
                'fund_type': 'bond',
            },
            {
                'mng_id':'B9F063',
                'mng_name':'邬传雁',
                'introduction':'现任泓德基金副总经理。风格稳健型，股债研究均深入，擅长股债择时，权益类增强',
                'fund_type': 'bond',
            }
        ]        
        self.result = pd.DataFrame(mng_info)

    def insert(self):
        self.append_data(ManagerInfoCollection.__tablename__, self.result)


import pandas as pd
import numpy as np
import time
from typing import Optional
from .calculator import Calculator
from ...constant import FundFactorType
from ...util.singleton import Singleton

START_DATE = '2005-01-01'
END_DATE = '2020-11-05'
BUCKET_NAME = 'tl-fund-factors'

class Factor(Calculator, metaclass=Singleton):
    
    def __init__(self, f_name: str, f_type: str=FundFactorType, f_level: str='basic'):
        self.f_name = f_name
        self.s3_file_name = f'{f_name}.parquet'
        self.f_type = f_type
        self.f_level = f_level
        self.s3_uri = f's3://{BUCKET_NAME}/{self.f_level}/' + self.s3_file_name
        self.factor: Optional[pd.DataFrame] = None

    def get(self) -> Optional[pd.DataFrame]:
        if self.factor is None:
            self.load()
        return self.factor

    def save(self) -> bool:
        if self.factor is None:
            return False
        self.factor.to_parquet(self.s3_uri, compression='gzip')
        print(f' upload to s3 success {self.f_name} {self.s3_uri}')
        return True

    def load(self):
        try:
            t0 = time.time()
            self.factor: pd.DataFrame = pd.read_parquet(self.s3_uri)
            t1 = time.time()
            print(f'\t[time][{self.f_level}] fetch factor: {self.f_name} from s3 cost {round(t1 - t0, 4)}')
        except Exception as e:
            print(f'retrieve data from s3 failed, (err_msg){e}; try to re-calc {self.f_name}')
            self.calc()
        return self.factor

    def data_reindex_daily_trading_day(self, data, tradingday):
        # 日线因子交易日对齐
        td = pd.to_datetime(data.index)
        td = [i.date() for i in td]
        data.index = td
        data = data.reindex(data.index.union(tradingday.index))
        data = data.ffill().reindex(tradingday.index)
        return data

    def data_reindex_daily_trading_day_not_fill(self, data, tradingday):
        td = pd.to_datetime(data.index)
        td = [i.date() for i in td]
        data.index = td
        data = data.reindex(data.index.union(tradingday.index))
        data = data.reindex(tradingday.index)
        return data

    def data_resample_monthly_ret(self, df, rule='1M', min_count=15):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    def data_resample_weekly_ret(self, df, rule='1W', min_count=3):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    def data_resample_monthly_nav(self, df, rule='1M'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    def data_resample_weekly_nav(self, df, rule='1W'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df



import pandas as pd
import numpy as np
import math
import json
import datetime
import time
from functools import partial
from typing import Optional, List, Union, Dict
from multiprocessing import Pool
from .base import Factor, START_DATE, END_DATE
from .basic_factors import *
from ...data.manager import DataManager
from ...data.struct import AssetTimeSpan
from ...data.manager.fund_info_filter import fund_info_update, get_active_fund_info
from ...constant import FundFactorType 

class FactorDataPrepare:

    def asset_indicator_with_benchmark_prepare(self):
        # load data
        fund_info = FundInfo().get()
        active_fund_info = ActiveFundInfo().get()
        index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        self.fund_ret = FundRetDailyModify().get()
        self.index_ret = IndexRetDaily().get()

        # prepare index_id : fund_list
        fund_to_index_dict = fund_info[fund_info.index_id.isin(index_list)][['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        index_fund = { index_id : [fund_idx for fund_idx, index_idx in fund_to_index_dict.items() if index_idx == index_id] for index_id in index_list}
        if 'active' in index_fund.keys():
            del index_fund['active']
        if 'conv_bond' in index_fund.keys():
            del index_fund['conv_bond']
        for index_id, fund_list in index_fund.items():
            active_list = active_fund_info[active_fund_info.index_id == index_id].fund_id
            fund_list = list(set(active_list).union(set(fund_list)).intersection(self.fund_ret.columns))
            index_fund[index_id] = fund_list

        # truncate data
        self.index_ret = self.index_ret[list(index_fund.keys())]
        self.index_fund = index_fund

    def indicator_fund_alive_prepare(self):
        self.fund_info = FundInfo().get()
        self.fund_nav = FundNavDailyModify().get()
        self.fund_start_date = self.fund_info.set_index('fund_id').to_dict()['start_date']
        self.fund_end_date = self.fund_info.set_index('fund_id').to_dict()['end_date']

    def indicator_with_benchmark_weekly_prepare(self):
        self.fund_info = FundInfo().get()
        self.fund_weekly_ret = FundRetWeeklyModify().get()
        self.index_weekly_ret = IndexRetWeekly().get().rename(columns=self.INDEX_DICT)[list(self.INDEX_DICT.values())]
        self.fund_wind_class_dic = self.fund_info.set_index('fund_id').to_dict()['wind_class_2']
        
    def mng_indicator_with_benchmark_weekly_prepare(self):
        self.mng_weekly_ret = ManagerIndexWeekly().get()
        self.index_weekly_ret = IndexRetWeekly().get().rename(columns=self.INDEX_DICT)[list(self.INDEX_DICT.values())]

    def indicator_with_daily_ret_prepare(self):
        self.fund_ret = FundRetDailyModify().get()

    def indicator_with_daily_nav_prepare(self):
        self.fund_nav = FundNavDailyModify().get()

    def indicator_with_monthly_nav_prepare(self):
        self.fund_nav = FundNavMonthlyModify().get()

    def indicator_with_monthly_ret_prepare(self):
        self.fund_ret = FundRetMonthlyModify().get()

    def fund_manager_size_prepare(self):
        self.trading_days_list = TradingDay().get().index
        self.fund_info = FundInfo().get()
        self.fund_manager_info = FundManagerInfo().get()
        self.fund_manager_info = self.fund_manager_info.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
        ## 计算基金经理指数时只去掉分级基金子基金， 不能回测交易的基金依然影响基金经理能力
        self.fund_info = self.fund_info[(self.fund_info.structure_type <= 1)
                                        & (~self.fund_info.wind_class_2.isnull())]
        self.fund_size = FundSize().get()
        self.fund_manager_info = self.fund_manager_info.set_index('fund_id')
        self.manager_name_dict = self.fund_manager_info.set_index('manager_id').to_dict()['name'] 

    def fund_manager_prepare(self):
        self.mng_index_list: Dict[str, pd.DataFrame] = {}  # 基金经理指数
        self.mng_best_fund: Dict[str, pd.DataFrame] = {}  # 基金经理代表作 
        self.trading_days_list = TradingDay().get().index
        self.full_trading_days_list = DataManager.basic_data(func_name='get_trading_day_list').datetime
        self.index_price = IndexCloseDaily().get().rename(columns=self.INDEX_DICT)
        self.index_ret = np.log(self.index_price).diff()
        self.fund_info = FundInfo().get()
        self.fund_manager_info = FundManagerInfo().get()
        self.fund_manager_info = self.fund_manager_info.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
        ## 计算基金经理指数时只去掉分级基金子基金， 不能回测交易的基金依然影响基金经理能力
        self.fund_info = self.fund_info[(self.fund_info.structure_type <= 1)
                                        & (~self.fund_info.wind_class_2.isnull())]

        self.fund_size = FundSizeCombine().get()
        self.fund_nav = FundNavDailyModify().get()
        self.fund_ret = FundRetDailyModify().get()
        self.manager_name_dict = self.fund_manager_info.set_index('manager_id').to_dict()['name'] 
        self.fund_manager_info = self.fund_manager_info.set_index('fund_id')
        # 每个基金经理人数   如果多人给惩罚
        fund_list = self.fund_manager_info.index.unique().tolist()
        self.fund_manager_num_df = self.mng_process_manager_num(fund_list, self.fund_ret.index)
        
        # 每个基金的规模  小于1e 给惩罚
        self.fund_manager_size_df = self.fund_size.copy()
        self.fund_manager_size_df[self.fund_manager_size_df < 1e8] = 0.6
        self.fund_manager_size_df[(self.fund_manager_size_df >= 1e8) & (self.fund_manager_size_df < 2e9)] = 0.7
        self.fund_manager_size_df[(self.fund_manager_size_df >= 2e9) & (self.fund_manager_size_df < 5e9)] = 0.8
        self.fund_manager_size_df[(self.fund_manager_size_df >= 5e9) & (self.fund_manager_size_df < 1e10)] = 0.9
        self.fund_manager_size_df[self.fund_manager_size_df >= 1e10] = 1
        self.fund_manager_size_df = self.fund_manager_size_df.fillna(self.PUNISH_RATIO)

        # 基金经理数据 根据被动主动分类
        pass_fund_id_list = self.fund_info[self.fund_info.wind_class_2.str.contains('被动')].fund_id.tolist()
        pass_fund_id_list = [i for i in pass_fund_id_list if i in self.fund_ret.columns] 
        acti_fund_id_list = [i for i in self.fund_ret.columns if i not in pass_fund_id_list] 
        _df_acti = pd.DataFrame(1, columns=acti_fund_id_list, index=self.fund_ret.index)
        _df_pass = pd.DataFrame(self.PUNISH_RATIO, columns=pass_fund_id_list, index=self.fund_ret.index)
        self.fund_manager_stype_df = pd.concat([_df_acti, _df_pass], axis=1)
        
        for fund_type in self.FUND_CLASSIFIER:
            self.mng_index_list[fund_type] = pd.DataFrame(index=self.fund_ret.index)
            self.mng_best_fund[fund_type] = pd.DataFrame(index=self.fund_ret.index)

    def mng_lambda_filter_date_range(self, single_fund_info: pd.DataFrame, date_index: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((date_index >= row.start) & (date_index <= row.end), index=date_index))
        return pd.concat(result, axis=1).sum(axis=1)

    def mng_process_fund_ret(self, this_manager_fund_ids, this_manager_info, date_index):
        res = []
        for fund_id in this_manager_fund_ids:
            single_fund_info = this_manager_info.loc[[fund_id]]
            res_i =  self.mng_lambda_filter_date_range(single_fund_info, date_index)
            res_i.name = fund_id
            res.append(res_i)
        return pd.concat(res, axis=1)

    def mng_process_manager_num(self, fund_id_list, date_index):
        res = []
        for fund_id in fund_id_list:
            fund_managers_info = self.fund_manager_info.loc[[fund_id]]
            fund_manager_list = fund_managers_info.manager_id.unique()
            _res = []
            for _mng_id in fund_manager_list:
                fund_manager_i = fund_managers_info[fund_managers_info['manager_id'] == _mng_id]
                single_fund_res = self.mng_lambda_filter_date_range(fund_manager_i, date_index)
                single_fund_res = pd.DataFrame(single_fund_res, columns=[_mng_id])
                _res.append(single_fund_res)
            manager_in_same_time = pd.concat(_res,axis=1)
            fund_manger_same_time = pd.DataFrame(manager_in_same_time.sum(axis=1), columns=[fund_id])            
            res.append(fund_manger_same_time)
        fund_manger_same_time = pd.concat(res, axis=1)
        fund_manger_same_time[fund_manger_same_time > 1] = self.PUNISH_RATIO
        return fund_manger_same_time

    def mng_process_manager_experience(self, fund_id_list, this_manager_info, date_index):
        res = []
        for fund_id in fund_id_list:
            single_fund_info = this_manager_info.loc[[fund_id]]
            work_day_df_i = self.mng_lambda_filter_date_range(single_fund_info, date_index)
            work_day_df_i.name = fund_id
            res.append(work_day_df_i)
        work_day_df = pd.DataFrame(res).T
        work_day_df = work_day_df.cumsum()
        low_cons = work_day_df < self.TRADING_DAYS_PER_YEAR
        mid_cons = (work_day_df >= self.TRADING_DAYS_PER_YEAR) & (work_day_df < 3 * self.TRADING_DAYS_PER_YEAR)
        hig_cons = work_day_df >= 3 * self.TRADING_DAYS_PER_YEAR
        work_day_df[low_cons] = self.HARD_PUNISH_RATIO
        work_day_df[mid_cons] = self.PUNISH_RATIO
        work_day_df[hig_cons] = 1
        return work_day_df

class UpdateDerivedFactorStart:
    pass

class FundAlive(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='FundAlive', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_fund_alive_prepare()
        df = pd.DataFrame(1,columns=self.fund_nav.columns,index=self.fund_nav.index)
        res = []
        for fund_id in df:
            s_d = self.fund_start_date[fund_id]
            e_d = self.fund_end_date[fund_id]
            df_i = df.loc[s_d:e_d][[fund_id]]
            res.append(df_i)
        self.factor = pd.concat(res,axis=1).sort_index()
        self.factor.index.name = 'datetime'

class TradeYear(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='TradeYear', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        self.factor = self.fund_nav.rolling(window=self.fund_nav.shape[0],min_periods=1).apply(self.rolling_trade_year, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())
        
class FeeRate(Factor):

    def __init__(self):
        super().__init__(f_name='FeeRate', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        fund_info = FundInfo().get()
        fee_dic = fund_info[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        self.factor = pd.DataFrame(fee_dic, columns=[datetime.date(2005,1,4)]).T
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AlphaBetaDaily3yI(Factor, FactorDataPrepare):
    # cost 20-30 mins
    def __init__(self):
        super().__init__(f_name='AlphaBetaDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        
    def loop_item(self, fund_id):
        df = self.fund_ret[[fund_id]].join(self.index_ret[[self.index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',self.index_id:'benchmark'}).reset_index()
        res = []
        pd.Series(df.index).rolling(
            window=self.time_range, min_periods=self.MIN_TIME_SPAN).apply(
            partial(self.rolling_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def calc(self, is_multi_cpu=False):
        self.asset_indicator_with_benchmark_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        # calculate
        result = []
        for self.index_id, fund_list in self.index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(self.loop_item, fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_id)
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()

class AlphaDaily3yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AlphaDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = AlphaBetaDaily3yI().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class BetaDaily3yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='BetaDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = AlphaBetaDaily3yI().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class TrackerrDaily3yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='TrackerrDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
    
    def loop_item(self, fund_id):
        df = self.fund_ret[[fund_id]].join(self.index_ret[[self.index_id]]).dropna()
        df['track_err'] = (df[fund_id] - df[self.index_id]).rolling(window=self.time_range, min_periods=self.MIN_TIME_SPAN).std(ddof=1)* np.sqrt(self.TRADING_DAYS_PER_YEAR)
        df = df[['track_err']].rename(columns={'track_err': fund_id})
        return df

    def calc(self, is_multi_cpu=False):
        self.asset_indicator_with_benchmark_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        
        # calculate
        result = []
        for self.index_id, fund_list in self.index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(self.loop_item, fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_id)
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class InfoRatioDaily3yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='InfoRatioDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
    
    def calc(self):
        alpha = AlphaDaily3yI().get()
        track_err = TrackerrDaily3yI().get()
        common_funds = alpha.columns.intersection(track_err.columns)
        self.factor = alpha[common_funds] / track_err[common_funds]
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MddDaily3y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MddDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        
        self.factor = self.fund_nav.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).apply(self.rolling_mdd, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class TreynorDaily3yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='TreynorDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        beta = BetaDaily3yI().get()
        self.indicator_with_daily_ret_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        common_funds = beta.columns.intersection(self.fund_ret.columns)
        fund_mean_ret = self.fund_ret.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).mean()
        fund_mean_ret_no_free_ret = fund_mean_ret - self.RISK_FEE_RATE_PER_DAY
        self.factor = fund_mean_ret_no_free_ret[common_funds] * self.TRADING_DAYS_PER_YEAR / beta[common_funds]
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class DownriskDaily3y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='DownriskDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_ret_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        down_ret = np.abs(np.minimum(self.fund_ret - self.RISK_FEE_RATE_PER_DAY, 0))
        self.factor = down_ret.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).mean()* np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class RetOverPeriodDaily3y(Factor, FactorDataPrepare):
    
    def __init__(self):
        super().__init__(f_name='RetOverPeriodDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        self.factor = self.fund_nav / self.fund_nav.fillna(method='bfill').shift(self.time_range) - 1
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualRetDaily3y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AnnualRetDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        self.factor = self.fund_nav.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).apply(self.rolling_annual_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualVolDaily3y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AnnualVolDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_ret_prepare()
        year = 3
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        self.factor = self.fund_ret.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class SharpeDaily3y(Factor):

    def __init__(self):
        super().__init__(f_name='SharpeDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        annual_ret = AnnualRetDaily3y().get()
        annual_vol = AnnualVolDaily3y().get()
        self.factor =  (annual_ret - self.RISK_FEE_RATE) / annual_vol
        self.factor = self.factor.replace(-np.Inf,np.nan).replace(np.Inf,np.nan)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AlphaBetaDaily1yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AlphaBetaDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        
    def loop_item(self, fund_id):
        df = self.fund_ret[[fund_id]].join(self.index_ret[[self.index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',self.index_id:'benchmark'}).reset_index()
        res = []
        pd.Series(df.index).rolling(
            window=self.time_range, min_periods=self.MIN_TIME_SPAN).apply(
            partial(self.rolling_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def calc(self, is_multi_cpu=False):
        self.asset_indicator_with_benchmark_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        # calculate
        result = []
        for self.index_id, fund_list in self.index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(self.loop_item, fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_id)
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
    
class AlphaDaily1yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AlphaDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = AlphaBetaDaily1yI().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class BetaDaily1yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='BetaDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = AlphaBetaDaily1yI().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class TrackerrDaily1yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='TrackerrDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
    
    def loop_item(self, fund_id):
        df = self.fund_ret[[fund_id]].join(self.index_ret[[self.index_id]]).dropna()
        df['track_err'] = (df[fund_id] - df[self.index_id]).rolling(window=self.time_range, min_periods=self.MIN_TIME_SPAN).std(ddof=1)* np.sqrt(self.TRADING_DAYS_PER_YEAR)
        df = df[['track_err']].rename(columns={'track_err': fund_id})
        return df

    def calc(self, is_multi_cpu=False):
        self.asset_indicator_with_benchmark_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        
        # calculate
        result = []
        for self.index_id, fund_list in self.index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(self.loop_item, fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_id)
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class InfoRatioDaily1yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='InfoRatioDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
    
    def calc(self):
        alpha = AlphaDaily1yI().get()
        track_err = TrackerrDaily1yI().get()
        common_funds = alpha.columns.intersection(track_err.columns)
        self.factor = alpha[common_funds] / track_err[common_funds]
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MddDaily1y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MddDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        
        self.factor = self.fund_nav.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).apply(self.rolling_mdd, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class TreynorDaily1yI(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='TreynorDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        beta = BetaDaily3yI().get()
        self.indicator_with_daily_ret_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        common_funds = beta.columns.intersection(self.fund_ret.columns)
        fund_mean_ret = self.fund_ret.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).mean()
        fund_mean_ret_no_free_ret = fund_mean_ret - self.RISK_FEE_RATE_PER_DAY
        self.factor = fund_mean_ret_no_free_ret[common_funds] * self.TRADING_DAYS_PER_YEAR / beta[common_funds]
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class DownriskDaily1y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='DownriskDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_ret_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        down_ret = np.abs(np.minimum(self.fund_ret - self.RISK_FEE_RATE_PER_DAY, 0))
        self.factor = down_ret.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).mean()* np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class RetOverPeriodDaily1y(Factor, FactorDataPrepare):
    
    def __init__(self):
        super().__init__(f_name='RetOverPeriodDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        self.factor = self.fund_nav / self.fund_nav.fillna(method='bfill').shift(self.time_range) - 1
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualRetDaily1y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AnnualRetDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        self.factor = self.fund_nav.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).apply(self.rolling_annual_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualVolDaily1y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AnnualVolDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_ret_prepare()
        year = 1
        self.time_range = year * self.TRADING_DAYS_PER_YEAR
        self.factor = self.fund_ret.rolling(window=self.time_range,min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class SharpeDaily1y(Factor):

    def __init__(self):
        super().__init__(f_name='SharpeDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        annual_ret = AnnualRetDaily1y().get()
        annual_vol = AnnualVolDaily1y().get()
        self.factor =  (annual_ret - self.RISK_FEE_RATE) / annual_vol
        self.factor = self.factor.replace(-np.Inf,np.nan).replace(np.Inf,np.nan)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualRetDailyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AnnualRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        self.factor = self.fund_nav.rolling(window=self.fund_nav.shape[0],min_periods=self.MIN_TIME_SPAN).apply(self.rolling_annual_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualVolDailyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='AnnualVolDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_ret_prepare()
        self.factor = self.fund_ret.rolling(window=self.fund_ret.shape[0],min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class AnnualRetMonthlyHalfY(Factor, FactorDataPrepare):
    # 不在日度做reindex 算ContinueRegValue用
    def __init__(self):
        super().__init__(f_name='AnnualRetMonthlyHalfY', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_monthly_nav_prepare()
        self.factor = self.fund_nav.rolling(window=6).apply(self.rolling_monthly_annual_ret, raw=True)
        self.factor = self.factor.dropna(axis=0, how='all')
        
class AnnualVolMonthlyHalfY(Factor, FactorDataPrepare):
    # 不在日度做reindex 算ContinueRegValue用
    def __init__(self):
        super().__init__(f_name='AnnualVolMonthlyHalfY', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_monthly_ret_prepare()
        self.factor = self.fund_ret.rolling(window=6).std(ddof=1) * np.sqrt(12)
        self.factor = self.factor.dropna(axis=0, how='all')

class SharpeMonthlyHalfY(Factor, FactorDataPrepare):
    # 不在日度做reindex 算ContinueRegValue用
    def __init__(self):
        super().__init__(f_name='SharpeMonthlyHalfY', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        annual_ret = AnnualRetMonthlyHalfY().get()
        annual_vol = AnnualVolMonthlyHalfY().get()
        self.factor =  (annual_ret - self.RISK_FEE_RATE) / annual_vol
        self.factor = self.factor.replace(-np.Inf,np.nan).replace(np.Inf,np.nan)
        self.factor = self.factor.dropna(axis=0, how='all')

class ContinueRegValue(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='ContinueRegValue', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        sharpe = SharpeMonthlyHalfY().get()
        self.factor = sharpe.rolling(window=sharpe.shape[0],min_periods=6).apply(self.rolling_auto_reg, raw=True)
        self.factor = self.factor.dropna(axis=0, how='all')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class TotalRetDailyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='TotalRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        self.factor = self.fund_nav.rolling(window=self.fund_nav.shape[0],min_periods=self.MIN_TIME_SPAN).apply(self.rolling_totol_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MddDailyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MddDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()        
        self.factor = self.fund_nav.rolling(window=self.fund_nav.shape[0],min_periods=self.MIN_TIME_SPAN).apply(self.rolling_mdd, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class RecentMonthRet(Factor, FactorDataPrepare):
    
    def __init__(self):
        super().__init__(f_name='RecentMonthRet', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_nav_prepare()
        self.factor = self.fund_nav.rolling(window=20,min_periods=5).apply(self.rolling_totol_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class DownsideStdDailyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='DownsideStdDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.indicator_with_daily_ret_prepare()
        self.factor = self.fund_ret.copy()
        self.factor[self.factor > 0] = 0
        self.factor = self.factor.rolling(window=self.factor.shape[0],min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundClAlphaBetaHistoryWeekly(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='FundClAlphaBetaHistoryWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def loop_item(self, fund_id):
        index_id = self.WIND_TYPE_DICT[self.fund_wind_class_dic[fund_id]]
        df = self.fund_weekly_ret[[fund_id]].join(self.index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=df.shape[0], min_periods=24).apply(
            partial(self.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def calc(self, is_multi_cpu=False):
        self.indicator_with_benchmark_weekly_prepare()
        fund_list = self.fund_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(self.loop_item, fund_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for fund_id in fund_list:
                res_i = self.loop_item(fund_id)
                if res_i is not None:
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
        self.factor.index.name = 'datetime'
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundClAlphaHistoryWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClAlphaHistoryWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = FundClAlphaBetaHistoryWeekly().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundClBetaHistoryWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClBetaHistoryWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = FundClAlphaBetaHistoryWeekly().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundClAlphaBeta1YWeekly(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='FundClAlphaBeta1YWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def loop_item(self, fund_id):
        index_id = self.WIND_TYPE_DICT[self.fund_wind_class_dic[fund_id]]
        df = self.fund_weekly_ret[[fund_id]].join(self.index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=52, min_periods=24).apply(
            partial(self.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def calc(self, is_multi_cpu=False):
        self.indicator_with_benchmark_weekly_prepare()
        fund_list = self.fund_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(self.loop_item, fund_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for fund_id in fund_list:
                res_i = self.loop_item(fund_id)
                if res_i is not None:
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
        self.factor.index.name = 'datetime'
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundClAlpha1YWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClAlpha1YWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = FundClAlphaBeta1YWeekly().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundClBeta1YWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClBeta1YWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = FundClAlphaBetaHistoryWeekly().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class ManagerIndex(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='ManagerIndex', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.fund_manager_prepare()
        fund_manager_list = list(self.manager_name_dict.keys())
        t0 = time.time()
        for m_id in fund_manager_list:
            _t0 = time.time()
            for fund_type in self.FUND_CLASSIFIER:
                idx = self.index_calculation(m_id, fund_type)
                if idx is not None:
                    self.mng_index_list[fund_type].loc[:, m_id] = idx # 该资产类别下记录计算结果
            _idx = fund_manager_list.index(m_id)
            _t1 = time.time()
            #print(f'm_id {m_id} {_idx} {len(fund_manager_list)} cost time {_t1 - _t0}')
        t1 = time.time()
        #print(f'total cost {t1 - t0}')
        for fund_type in self.FUND_CLASSIFIER:
            self.mng_index_list[fund_type] = self.mng_index_list[fund_type].replace(0, np.nan).dropna(how='all')
    
        res = []
        for fund_type, mng_index in self.mng_index_list.items():
            df = mng_index.copy()
            df.columns = [fund_type + '_' + i for i in df.columns]
            res.append(df)
        self.mng_index = pd.concat(res,axis=1).dropna(axis=1,how='all')
          
        res = []
        for fund_type, mng_index in self.mng_best_fund.items():
            df = mng_index.copy()
            print(fund_type)
            df.columns = ['bf_' + fund_type + '_' + i for i in df.columns] 
            res.append(df)
        self.mng_best_fund = pd.concat(res,axis=1).dropna(axis=1,how='all')
        # self.s3_uri = 's3://tl-fund-factors/derived/ManagerBestFund.parquet'
        # self.save()
        
    def save(self):
        mng_index_uri = self.s3_uri
        self.mng_index.to_parquet(mng_index_uri, compression='gzip')
        print(f' upload to s3 success {self.f_name} {mng_index_uri}')

        mng_best_fund_uri = self.s3_uri.replace('ManagerIndex','ManagerBestFund')
        self.mng_best_fund.to_parquet(mng_best_fund_uri, compression='gzip')
        print(f' upload to s3 success {self.f_name} {mng_best_fund_uri}')

    def index_calculation(self, m_id: str, fund_type: str):
        wind_class_2_list = self.FUND_CLASSIFIER[fund_type]
        # 该基金经理管理基金信息
        this_manager_info = self.fund_manager_info[(self.fund_manager_info['wind_class_2'].isin(wind_class_2_list))
                            & (self.fund_manager_info['manager_id'] == m_id)]
        if this_manager_info.shape[0] == 0:
            return 
        index_start_date = this_manager_info.start.min()
        index_end_date = this_manager_info.end.max()
        # 该基金经理管理的同类基金列表
        this_manager_fund_ids = this_manager_info.index.unique().tolist()
        _fund_id_list1 = self.fund_ret.columns.tolist()
        _fund_id_list2 = self.fund_manager_size_df.columns.tolist()
        this_manager_fund_ids = list(set(this_manager_fund_ids).intersection(_fund_id_list1).intersection(_fund_id_list2))
        if len(this_manager_fund_ids) == 0:
            return
        fund_ret = self.fund_ret.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 非管理时间，基金收益赋值0
        manager_status = self.mng_process_fund_ret(this_manager_fund_ids, this_manager_info, fund_ret.index)
        fund_ret = fund_ret * manager_status
        # 基金经理管理的基金 人数状态 单人为 1 ，多人为0.8
        manager_num = self.fund_manager_num_df.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 基金经理管理的基金 规模状态  > 100e : 1, > 50e : 0.9, > 20e : 0.8, > 1e : 0.7, < 1e : 0.6
        manager_size = self.fund_manager_size_df.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 基金经理管理的基金 负责年限 大于3 1， 大于1 0.8， 小于1 0.6
        manager_year = self.mng_process_manager_experience(this_manager_fund_ids, this_manager_info, fund_ret.index)
        # 基金经理管理的基金 主动 1 , 被动 0.8
        manager_style = self.fund_manager_stype_df.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 权重累乘
        fund_weight = manager_num * manager_size * manager_year * manager_style * manager_status
        # 权重和归1
        fund_weight['sum'] = fund_weight.sum(axis=1).T
        fund_weight = fund_weight[this_manager_fund_ids].divide(fund_weight['sum'], axis=0)
        # 基金经理 收益
        total_ret = (fund_ret * fund_weight).sum(axis=1)
        total_ret.name = m_id
        total_ret = pd.DataFrame(total_ret)
        if total_ret.shape[0] == 0:
            return 
        # 基金经理 收益 用指数填充待业期
        index_date_list = manager_status.index[manager_status.sum(axis=1) < 1]
        total_ret.loc[index_date_list, m_id] = self.index_ret.loc[index_date_list, fund_type]
        # 收益初值赋0
        last_trading_day = self.full_trading_days_list[self.full_trading_days_list < total_ret.index.values[0]].values[-1]
        total_ret.loc[last_trading_day] = 0
        total_ret = total_ret.sort_index()
        # 做指数
        fund_manager_index = np.exp(total_ret.cumsum())
        # 基金经理代表作
        manager_best_fund = pd.DataFrame((fund_weight + manager_status.cumsum()/1e8).idxmax(axis=1), columns=[m_id])
        self.mng_best_fund[fund_type].loc[:, m_id] = manager_best_fund
        return fund_manager_index

class ManagerBestFund(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='ManagerBestFund', f_type=FundFactorType.DERIVED, f_level='derived')

class ManagerIndexRetDaily(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexRetDaily', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = np.log(mng_index).diff(1)
        self.factor = self.data_reindex_daily_trading_day_not_fill(self.factor, TradingDay().get())

class ManagerIndexWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = self.data_resample_weekly_nav(mng_index).dropna(axis=0, how='all')

class ManagerIndexRetWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexWeekly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndexWeekly().get()
        self.factor = np.log(mng_index).diff(1)

class ManagerIndexMonthly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexMonthly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = self.data_resample_monthly_nav(mng_index).dropna(axis=0, how='all')

class ManagerIndexRetMonthly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexRetMonthly', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndexMonthly().get()
        self.factor = np.log(mng_index).diff(1)

class MngAnnualRetDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = mng_index.rolling(window=mng_index.shape[0],min_periods=self.MIN_TIME_SPAN).apply(self.rolling_annual_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngAnnualRetDaily1Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualRetDaily1Y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = mng_index.rolling(window=self.TRADING_DAYS_PER_YEAR,min_periods=self.MIN_TIME_SPAN).apply(self.rolling_annual_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngAnnualRetDaily3Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualRetDaily3Y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = mng_index.rolling(window=self.TRADING_DAYS_PER_YEAR*3,min_periods=self.MIN_TIME_SPAN).apply(self.rolling_annual_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())


class MngTotalRetDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngTotalRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = mng_index.rolling(window=mng_index.shape[0],min_periods=self.MIN_TIME_SPAN).apply(self.rolling_totol_ret, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngMddDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngMddDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = mng_index.rolling(window=mng_index.shape[0],min_periods=self.MIN_TIME_SPAN).apply(self.rolling_mdd, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngAnnualVolDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualVolDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index_ret = ManagerIndexRetDaily().get()
        self.factor = mng_index_ret.rolling(window=mng_index_ret.shape[0],min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngDownsideStdDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngDownsideStdDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index_ret = ManagerIndexRetDaily().get().copy()
        mng_index_ret[mng_index_ret > 0] = 0
        self.factor = mng_index_ret.rolling(window=mng_index_ret.shape[0],min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngDownsideStdMonthlyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngDownsideStdMonthlyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index_ret = ManagerIndexRetMonthly().get().copy()
        mng_index_ret[mng_index_ret > 0] = 0
        self.factor = mng_index_ret.rolling(window=mng_index_ret.shape[0],min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngFundTypeTradingDays(Factor):

    def __init__(self):
        super().__init__(f_name='MngFundTypeTradingDays', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        mng_index = ManagerIndex().get()
        self.factor = mng_index.rolling(window=mng_index.shape[0],min_periods=1).apply(self.rolling_trade_year, raw=True)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngClAlphaBetaWeekly1Y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngClAlphaBetaWeekly1Y', f_type=FundFactorType.DERIVED, f_level='derived')

    def loop_item(self, mng_id):
        index_id = mng_id.split('_')[0]
        df = self.mng_weekly_ret[[mng_id]].join(self.index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={mng_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=52, min_periods=24).apply(
            partial(self.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{mng_id}', 'beta':f'beta_{mng_id}'})
        return df

    def calc(self, is_multi_cpu=False):
        self.mng_indicator_with_benchmark_weekly_prepare()
        mng_list = self.mng_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(self.loop_item, mng_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for mng_id in mng_list:
                res_i = self.loop_item(mng_id)
                if res_i is not None:
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
        self.factor.index.name = 'datetime'
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngClAlphaWeekly1Y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngClAlphaWeekly1Y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = MngClAlphaBetaWeekly1Y().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.replace('alpha_','') for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngClBetaWeekly1Y(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngClBetaWeekly1Y', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = MngClAlphaBetaWeekly1Y().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.replace('beta_','') for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngClAlphaBetaWeeklyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngClAlphaBetaWeeklyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def loop_item(self, mng_id):
        index_id = mng_id.split('_')[0]
        df = self.mng_weekly_ret[[mng_id]].join(self.index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={mng_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=df.shape[0], min_periods=24).apply(
            partial(self.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{mng_id}', 'beta':f'beta_{mng_id}'})
        return df

    def calc(self, is_multi_cpu=False):
        self.mng_indicator_with_benchmark_weekly_prepare()
        mng_list = self.mng_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(self.loop_item, mng_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for mng_id in mng_list:
                res_i = self.loop_item(mng_id)
                if res_i is not None:
                    result.append(res_i)
        self.factor = pd.concat(result, axis=1)
        self.factor = self.factor.sort_index()
        self.factor.index.name = 'datetime'
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngClAlphaWeeklyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngClAlphaWeeklyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = MngClAlphaBetaWeeklyHistory().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.replace('alpha_','') for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngClBetaWeeklyHistory(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngClBetaWeeklyHistory', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.factor = MngClAlphaBetaWeeklyHistory().get()
        col_name = self.factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self.factor = self.factor[factor_list]
        name_dic = { i : i.replace('beta_','') for i in factor_list}
        self.factor = self.factor[factor_list].rename(columns=name_dic)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class MngFundSize(Factor, FactorDataPrepare):

    def __init__(self):
        super().__init__(f_name='MngFundSize', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        self.fund_manager_size_prepare()
        fund_manager_list = list(self.manager_name_dict.keys())
        t0 = time.time()
        loop_paras = []
        for m_id in fund_manager_list:
            for fund_type in self.FUND_CLASSIFIER:
                loop_paras.append([m_id, fund_type])
        result = []
        for para_i in loop_paras:
            df = self.size_calculation(para_i) 
            result.append(df)
        t1 = time.time()
        print(f'cost time {t1 - t0}')
        self.factor = pd.concat(result,axis=1).dropna(axis=1,how='all')

    def size_calculation(self, para_i):
        m_id = para_i[0]
        fund_type = para_i[1]
        wind_class_2_list = self.FUND_CLASSIFIER[fund_type]
        this_manager_info = self.fund_manager_info[(self.fund_manager_info['wind_class_2'].isin(wind_class_2_list))
                            & (self.fund_manager_info['manager_id'] == m_id)]
        if this_manager_info.shape[0] == 0:
            return None
        index_start_date = this_manager_info.start.min()
        index_end_date = this_manager_info.end.max()
        this_manager_fund_ids = this_manager_info.index.unique().intersection(self.fund_size.columns).tolist()
        if len(this_manager_fund_ids) < 1:
            return None
        fund_size = self.fund_size.loc[index_start_date:index_end_date, this_manager_fund_ids]
        manager_status = self.mng_process_fund_ret(this_manager_fund_ids, this_manager_info, fund_size.index)
        fund_size = fund_size * manager_status
        fund_size = pd.DataFrame(fund_size.sum(axis=1), columns=[f'{fund_type}_{m_id}'])
        return fund_size

class UpdateDerivedFactorEnd:
    pass


import dataclasses
import datetime
import pandas as pd
import numpy as np
import math
import statsmodels.api as sm
from statsmodels import regression
from scipy.stats import gmean
from functools import partial
import statsmodels.api as sm
from statsmodels.tsa.ar_model import AutoReg
from scipy.stats import kurtosis, skew

class CalculatorBase:

    TRADING_DAYS_PER_YEAR = 242
    TOTAL_DAYS_PER_YEAR = 365

    @staticmethod
    def calc_cl_alpha_beta(total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1]}
    
    @staticmethod
    def calc_hm_alpha_beta(total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 1][X[:, 1] < 0] = 0
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                }

    @staticmethod
    def calc_continue_regress_v(monthly_ret, risk_free_rate):
        # 半年周期
        window = 6
        yearly_multiplier = 2
        period_num = 12
        annual_ret = monthly_ret.rolling(window=window).apply(lambda x : x.sum() * yearly_multiplier - risk_free_rate)
        annual_vol = monthly_ret.rolling(window=window).apply(lambda x : x.std(ddof=1) * np.sqrt(period_num))
        sharpe = (annual_ret / annual_vol)
        sharpe = sharpe.replace(np.Inf, np.nan).replace(-np.Inf, sharpe).dropna() # 空值sharpe赋值np.nan
        if sharpe.shape[0] < 6:
            return np.nan
        mod = AutoReg(endog=sharpe.values, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    @staticmethod
    def get_stat_result(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        benchmark_values:  pd.core.series.Series=None,
                        risk_free_rate:    float=0.025,
                        frequency:         str='1D'):# '1W' '1M'
        # TODO resample weekly monthly        
        risk_free_per_day = risk_free_rate / CalculatorBase.TRADING_DAYS_PER_YEAR
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        assert len(dates) == len(values), 'date and port_values has different numbers'
        sr = pd.Series(values, index=dates).sort_index().dropna()
        sr_ret = np.log(sr).diff(1).iloc[1:]
        start_date = sr.index[0] 
        end_date = sr.index[-1]
        days = (dates[-1] - dates[0]).days
        last_unit_nav = sr[-1] / sr[0]
        cumu_ret = last_unit_nav - 1
        annual_ret = math.pow(last_unit_nav, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1
        annual_vol = sr_ret.std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
        recent_1m_ret = sr[-1] / sr[-20] - 1 if len(sr) > 20 else sr[-1] / sr[0] - 1
        recent_3m_ret = sr[-1] / sr[-60] - 1 if len(sr) > 60 else sr[-1] / sr[0] - 1
        recent_6m_ret = sr[-1] / sr[-120] - 1 if len(sr) > 120 else sr[-1] / sr[0] - 1
        recent_5y_ret = sr[-1] / sr[-CalculatorBase.TRADING_DAYS_PER_YEAR*5] -1 if len(sr) > CalculatorBase.TRADING_DAYS_PER_YEAR*5 else sr[-1] / sr[0] - 1
        worst_3m_ret = sr.pct_change(60).min()
        worst_6m_ret = sr.pct_change(120).min()
        last_mv_diff = sr[-1] - sr[-2]
        sharpe = (annual_ret - risk_free_rate) / annual_vol
        recent_drawdown = 1 - (sr[-1] / sr.max())
        if np.isnan(recent_drawdown):
            recent_mdd_date1 = None
            recent_mdd_lens = None
        else:
            recent_mdd_date1 = sr.idxmax()
            recent_mdd_lens = (end_date - recent_mdd_date1).days
        mdd_part =  sr / sr.rolling(window=sr.shape[0], min_periods=1).max()
        mdd = 1 - mdd_part.min()
        if np.isnan(mdd):
            mdd_date2 = None
            mdd_date1 = None
            mdd_lens = None
        else:
            mdd_date2 = mdd_part.idxmin()
            mdd_date1 = sr[:mdd_date2].idxmax()
            mdd_lens = (mdd_date2-mdd_date1).days
        calmar = (annual_ret - risk_free_rate) / mdd
        sr_ret_no_risk_free = sr_ret - risk_free_rate / CalculatorBase.TRADING_DAYS_PER_YEAR
        downside_risk = np.minimum(sr_ret_no_risk_free,0).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
        sortino = (annual_ret - risk_free_rate) / downside_risk
        var =  np.quantile(sr_ret, 0.05) # 按照最大回撤处理， 损失非负， 计算ervar用
        cvar = max(0, -sr_ret[sr_ret < var].mean())
        var = - var
        ervar = (annual_ret - risk_free_rate) / var
        skew_value = skew(sr_ret)
        kurtosis_value = kurtosis(sr_ret)        
        
        if benchmark_values is not None:
            assert len(dates) == len(benchmark_values), 'date and bench_values has different numbers'
            sr_benchmark = pd.Series(benchmark_values, index=dates).sort_index().dropna()
            sr_benchmark_ret = np.log(sr_benchmark).diff(1).iloc[1:]
            last_unit_benchmark = sr_benchmark[-1] / sr_benchmark[0]
            annual_ret_benchmark = math.pow(last_unit_benchmark, CalculatorBase.TOTAL_DAYS_PER_YEAR / days) - 1
            df = pd.concat([sr, sr_benchmark],axis=1)
            df.columns= ['fund_ret','index_ret']
            df_monthly = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule='1M').last()
            df_monthly.index = df_monthly.index.date
            monthly_ret = df_monthly.pct_change(1).dropna()
            continue_value_m = CalculatorBase.calc_continue_regress_v(monthly_ret.fund_ret,risk_free_rate)
            df_ret_risk_free = (df.pct_change(1) - risk_free_per_day).dropna()
            cl_res = CalculatorBase.calc_cl_alpha_beta(df_ret_risk_free.to_numpy())
            alpha_cl = cl_res['alpha']
            beta_cl = cl_res['beta']
            hm_res = CalculatorBase.calc_hm_alpha_beta(df_ret_risk_free.to_numpy())
            alpha_hm = hm_res['alpha']
            beta_hm = hm_res['beta']
            excess_ret = annual_ret - annual_ret_benchmark
            win_rate = np.mean((sr_ret > sr_benchmark_ret) * 1)
            corr = sr_ret.corr(sr_benchmark_ret)    
            track_err = (sr_ret - sr_benchmark_ret).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
            info = excess_ret / track_err
            x = sr_benchmark_ret.dropna().values
            y = sr_ret.dropna().values
            x = sm.add_constant(x)
            model = regression.linear_model.OLS(y,x).fit()
            alpha = model.params[0]
            beta = model.params[1]
            treynor = (annual_ret - risk_free_rate) / beta
            not_system_risk = sum((sr_ret - risk_free_per_day)**2) \
                             - alpha * sum(sr_ret - risk_free_per_day) \
                             - beta * sum((sr_ret - risk_free_per_day) * (sr_benchmark_ret - risk_free_per_day))
            not_system_risk = np.sqrt(not_system_risk / (len(sr_ret) - 2))
            alpha = alpha * CalculatorBase.TRADING_DAYS_PER_YEAR
            df_ret = np.log(df).diff(1).iloc[1:]
            df_ret_up = df_ret[df_ret.index_ret > 0] + 1 
            up_cap_top = df_ret_up.fund_ret.product() ** (CalculatorBase.TRADING_DAYS_PER_YEAR/len(df_ret_up)) - 1
            up_cap_bot = df_ret_up.index_ret.product() ** (CalculatorBase.TRADING_DAYS_PER_YEAR/len(df_ret_up)) - 1
            up_capture = up_cap_top / up_cap_bot

            df_ret_down = df_ret[df_ret.index_ret < 0] + 1
            down_cap_top = df_ret_down.fund_ret.product() ** (CalculatorBase.TRADING_DAYS_PER_YEAR/len(df_ret_down)) - 1
            down_cap_bot = df_ret_down.index_ret.product() ** (CalculatorBase.TRADING_DAYS_PER_YEAR/len(df_ret_down)) - 1
            down_capture = down_cap_top / down_cap_bot

            df_ret['excess_ret'] = df_ret.fund_ret - df_ret.index_ret
            mod = AutoReg(endog=df_ret.excess_ret.values, lags=1)
            res = mod.fit()
            excess_ret_cotinue = res.params[0]
            return {
                'start_date':start_date,
                'end_date':end_date,
                'last_unit_nav':last_unit_nav,
                'cumu_ret':cumu_ret,
                'annual_ret':annual_ret,
                'annual_vol':annual_vol,
                'sharpe':sharpe,
                'recent_1m_ret':recent_1m_ret,
                'recent_3m_ret':recent_3m_ret,
                'recent_6m_ret':recent_6m_ret,
                'recent_5y_ret':recent_5y_ret,
                'worst_3m_ret':worst_3m_ret,
                'worst_6m_ret':worst_6m_ret,
                'last_mv_diff':last_mv_diff,
                'recent_drawdown':recent_drawdown,
                'recent_mdd_date1':recent_mdd_date1,
                'recent_mdd_lens':recent_mdd_lens,
                'mdd':mdd,
                'mdd_date1':mdd_date1,
                'mdd_date2':mdd_date2,
                'mdd_lens':mdd_lens,
                'calmar':calmar,
                'downside_risk':downside_risk,
                'sortino':sortino,
                'var':var,
                'ervar':ervar,
                'skew':skew_value,
                'kurtosis':kurtosis_value,
                'alpha_cl':alpha_cl,
                'beta_cl':beta_cl,
                'alpha_hm':alpha_hm,
                'beta_hm':beta_hm,
                'excess_ret':excess_ret,
                'win_rate':win_rate,
                'corr':corr,
                'track_err':track_err,
                'info':info,
                'alpha':alpha,
                'beta':beta,
                'treynor':treynor,
                'not_system_risk':not_system_risk,
                'cvar':cvar,
                'up_capture':up_capture,
                'down_capture':down_capture,
                'excess_ret_cotinue':excess_ret_cotinue,
                'continue_value_m':continue_value_m,
            }
        else:
            return {
                'start_date':start_date,
                'end_date':end_date,
                'last_unit_nav':last_unit_nav,
                'cumu_ret':cumu_ret,
                'annual_ret':annual_ret,
                'annual_vol':annual_vol,
                'sharpe':sharpe,
                'recent_1m_ret':recent_1m_ret,
                'recent_3m_ret':recent_3m_ret,
                'recent_6m_ret':recent_6m_ret,
                'recent_5y_ret':recent_5y_ret,
                'worst_3m_ret':worst_3m_ret,
                'worst_6m_ret':worst_6m_ret,
                'last_mv_diff':last_mv_diff,
                'recent_drawdown':recent_drawdown,
                'recent_mdd_date1':recent_mdd_date1,
                'recent_mdd_lens':recent_mdd_lens,
                'mdd':mdd,
                'mdd_date1':mdd_date1,
                'mdd_date2':mdd_date2,
                'mdd_lens':mdd_lens,
                'calmar':calmar,
                'downside_risk':downside_risk,
                'sortino':sortino,
                'var':var,
                'cvar':cvar,
                'ervar':ervar,
                'skew':skew_value,
                'kurtosis':kurtosis_value,
            }
            




import os
from ...util.config import SurfingConfigurator

# 在to_parquet/read_parquet中通过storage_options传递如下参数的方法不好用，这里直接设置环境变量
conf = SurfingConfigurator().get_aws_settings()
os.environ['AWS_ACCESS_KEY_ID'] = conf.aws_access_key_id
os.environ['AWS_SECRET_ACCESS_KEY'] = conf.aws_secret_access_key
os.environ['AWS_DEFAULT_REGION'] = conf.region_name



import pandas as pd
import numpy as np
import math
import json
import re
from typing import Optional, List, Union, Dict
from .base import Factor, START_DATE, END_DATE
from ...data.manager import DataManager
from ...data.manager.fund_info_filter import fund_info_update, get_active_fund_info, get_conv_funds, filter_fund_info
from ...constant import FundFactorType 
from ..struct import AssetTimeSpan
class UpdateBasicFactorStart:
    pass

'''

monthly data
scale adjusted
'''

class FundInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_info')
        self.factor = fund_info_update(self.factor).drop(columns=['_update_time'], errors='ignore')

class InMarketFundInfo(Factor):

    def __init__(self):
        super().__init__(f_name='InMarketFundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):#TODO
        self.factor = DataManager.basic_data(func_name='get_fund_status').drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor[~self.factor['trade_status'].isnull()]
        
class CloseFundInfo(Factor):

    def __init__(self):
        super().__init__(f_name='CloseFundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_open_info').drop(columns=['_update_time'], errors='ignore')

class FilteredFundInfoI(Factor):
    
    def __init__(self):
        super().__init__(f_name='FilteredFundInfoI', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundInfo().get()
        index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        self.factor = filter_fund_info(self.factor, index_list)
        in_market_fund_info = InMarketFundInfo().get()
        close_fund_info = CloseFundInfo().get()
        self.factor = self.factor[(~self.factor.fund_id.isin(in_market_fund_info.fund_id))
                                & (~self.factor.fund_id.isin(close_fund_info.fund_id))]

class TradingDay(Factor):

    def __init__(self):
        super().__init__(f_name='TradingDay', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_trading_day_list').set_index('datetime').drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor[(self.factor.index >= pd.to_datetime(START_DATE).date())
                                 & (self.factor.index <= pd.to_datetime(END_DATE).date())]

class IndexInfo(Factor):

    def __init__(self):
        super().__init__(f_name='IndexInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_index_info').drop(columns=['_update_time'], errors='ignore')

class FundManagerInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundManagerInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.derived_data(func_name='get_fund_manager_info').drop(columns=['_update_time'], errors='ignore')
        fund_info = FundInfo().get()
        self.factor = pd.merge(self.factor, fund_info[['fund_id','company_id','wind_class_2']], how='left', on='fund_id' )

class FundConvInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundConvInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        fund_conv_stats = DataManager.basic_data(func_name='get_fund_conv_stats')
        fund_info = FundInfo().get()
        self.factor = get_conv_funds(fund_conv_stats, fund_info)
        self.factor = pd.DataFrame(self.factor,columns=['fund_id'])
        
class FundBenchmarkInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundBenchmarkInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_benchmark').drop(columns=['_update_time'], errors='ignore')

class FilteredFundInfoWhole(Factor):

    def __init__(self):
        super().__init__(f_name='FilteredFundInfoWhole', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        fund_info = FundInfo().get()
        benchmark_info = FundBenchmarkInfo().get()
        index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        fund_info_index = filter_fund_info(fund_info, index_list)
        fund_info_active = get_active_fund_info(fund_info, benchmark_info)
        in_market_fund_info = InMarketFundInfo().get()
        close_fund_info = CloseFundInfo().get()
        fund_info = fund_info[(fund_info.fund_id.isin(fund_info_active.fund_id))
                            | (fund_info.fund_id.isin(fund_info_index.fund_id))]
        self.factor = fund_info[(~fund_info.fund_id.isin(in_market_fund_info.fund_id))
                                & (~fund_info.fund_id.isin(close_fund_info.fund_id))]

class ActiveFundInfo(Factor): #based on fund_info fund_benchmark_info

    def __init__(self):
        super().__init__(f_name='ActiveFundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = get_active_fund_info(FundInfo().get(), FundBenchmarkInfo().get()).drop(columns=['_update_time'], errors='ignore')

class IndexCloseDaily(Factor):

    def __init__(self):
        super().__init__(f_name='IndexCloseDaily', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_index_price_dt', start_date=START_DATE, columns=['close']).drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor.pivot(index='datetime',columns='index_id',values='close')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class IndexRetDaily(Factor): # based on index_close_daily

    def __init__(self):
        super().__init__(f_name='IndexRetDaily', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = IndexCloseDaily().get()
        self.factor = np.log(self.factor).diff(1)
        self.factor = self.factor.replace(np.Inf,np.nan).replace(-np.Inf,np.nan)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class IndexRetWeekly(Factor): # based on IndexRetDaily

    def __init__(self):
        super().__init__(f_name='IndexRetWeekly', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = IndexRetDaily().get()
        self.factor = self.data_resample_weekly_ret(self.factor).dropna(axis=0, how='all')

class IndexRetMonthly(Factor): # based on IndexRetDaily

    def __init__(self):
        super().__init__(f_name='IndexRetMonthly', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = IndexRetDaily().get()
        self.factor = self.data_resample_monthly_ret(self.factor).dropna(axis=0, how='all')

class FundNavDaily(Factor):

    def __init__(self):
        super().__init__(f_name='FundNavDaily', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_nav_with_date', start_date=START_DATE).drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor.pivot(index='datetime',columns='fund_id',values='adjusted_net_value')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundRetDaily(Factor): # based on FundNavDaily

    def __init__(self):
        super().__init__(f_name='FundRetDaily', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundNavDaily().get()
        self.factor = np.log(self.factor).diff(1)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundSize(Factor):

    def __init__(self):
        super().__init__(f_name='FundSize', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_size_range', start_date=START_DATE).drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor.pivot(index='datetime',columns='fund_id',values='size')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundSizeCombine(Factor): # based on FundSize

    def __init__(self):
        super().__init__(f_name='FundSizeCombine', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        fund_info = FundInfo().get()
        fund_size = FundSize().get()
        fund_info['trim_name'] = fund_info.desc_name.map(lambda x : re.subn(r'[ABCDEFR]{1,2}(\(人民币\)|\(美元现汇\)|\(美元现钞\)|1|2|3)?$', '', x)[0])
        trim_name_df = fund_info.set_index('fund_id')[['trim_name']]
        fund_id_df = fund_info.set_index('trim_name')[['fund_id']]
        self.factor = fund_size.T.join(trim_name_df).groupby('trim_name').sum()
        self.factor = self.factor.join(fund_id_df).set_index('fund_id').T
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundCompanyHold(Factor):

    def __init__(self):
        super().__init__(f_name='FundCompanyHold', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_company_hold', start_date=START_DATE).drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor.pivot(index='datetime',columns='fund_id',values='institution_holds')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundPersonHold(Factor):

    def __init__(self):
        super().__init__(f_name='FundPersonHold', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_fund_persional_hold', start_date=START_DATE).drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor.pivot(index='datetime',columns='fund_id',values='personal_holds')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundRetDailyModify(Factor): # based on fund_info fund_ret_daily

    def __init__(self):
        super().__init__(f_name='FundRetDailyModify', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        fund_info = FundInfo().get()
        fund_ret = FundRetDaily().get()
        _res = []
        _fund_ret_columns = fund_ret.columns.tolist()
        for fund_type, filter_ratio in self.FILTER_NAV_RATIO.items():
            _info_fund_list = fund_info[fund_info['wind_class_2'].isin(self.FUND_CLASSIFIER[fund_type])].fund_id.tolist()
            _fund_list = list(set(_info_fund_list).intersection(_fund_ret_columns))
            _fund_ret = fund_ret[_fund_list].copy()
            _fund_ret[_fund_ret > filter_ratio] = 0
            _res.append(_fund_ret)
        self.factor = pd.concat(_res,axis=1)
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())
        self.factor = self.factor.loc[:,~self.factor.columns.duplicated()].copy()

class FundRetWeeklyModify(Factor): # based on FundRetDailyModify

    def __init__(self):
        super().__init__(f_name='FundRetWeeklyModify', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundRetDailyModify().get()
        self.factor = self.data_resample_weekly_ret(self.factor).dropna(axis=0, how='all')

class FundRetMonthlyModify(Factor): # based on FundRetDailyModify

    def __init__(self):
        super().__init__(f_name='FundRetMonthlyModify', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundRetDailyModify().get()
        self.factor = self.data_resample_monthly_ret(self.factor).dropna(axis=0, how='all')

class FundNavDailyModify(Factor): # based on fund_ret_daily_modify

    def __init__(self):
        super().__init__(f_name='FundNavDailyModify', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundRetDailyModify().get()
        self.factor = np.exp(self.factor.cumsum())
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundNavWeeklyModify(Factor): # based on FundNavDailyModify

    def __init__(self):
        super().__init__(f_name='FundNavWeeklyModify', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundNavDailyModify().get()
        self.factor = self.data_resample_weekly_nav(self.factor).dropna(axis=0, how='all')

class FundNavMonthlyModify(Factor): # based on FundNavDailyModify

    def __init__(self):
        super().__init__(f_name='FundNavMonthlyModify', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundNavDailyModify().get()
        self.factor = self.data_resample_monthly_nav(self.factor).dropna(axis=0, how='all')

class Macroeco(Factor):

    def __init__(self):
        super().__init__(f_name='Macroeco', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.raw_data(func_name='get_em_macroeconomic_daily', codes=['EX_DR_RATIO']).drop(columns='_update_time')
        self.factor = self.factor.pivot_table(index='datetime', columns='codes', values='value')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundBenchmarkRet(Factor): # based on fund_benchmark_info macroeco index_close_daily fund_info fund_nav_daily fund_info

    def __init__(self):
        super().__init__(f_name='FundBenchmarkRet', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        fund_benchmark_info = FundBenchmarkInfo().get()
        macroeco = Macroeco().get()
        index_price = IndexCloseDaily().get()
        fund_info = FundInfo().get()
        fund_nav = FundNavDaily().get()
        wind_class_2_dict = fund_info.set_index('fund_id')['wind_class_2'].to_dict()

        benchmark_list: Dict[str, float] = {}
        # 遍历每一只基金的benchmark进行处理
        for row in fund_benchmark_info.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index == '1':
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index in ('ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            break
                        else:
                            try:
                                if index == 'nonor':
                                    # 在这里我们用超额存款准备金率替代同业存款利率
                                    ra: pd.Series = macroeco.loc[:, 'EX_DR_RATIO']
                                else:
                                    ra: pd.Series = index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                break
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight / self.TRADING_DAYS_PER_YEAR)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            break
                        else:
                            try:
                                ra: pd.Series = index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                break
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
                else:
                    if values or cons:
                        the_sum: float = sum(values)
                        if cons:
                            the_sum += np.log(math.pow(1 + cons, 1 / self.TRADING_DAYS_PER_YEAR))
                        benchmark_list[row.fund_id] = the_sum
        self.factor =  pd.DataFrame.from_dict(benchmark_list)
        benchmark_cols = self.factor.columns.tolist()
        fund_nav_cols = fund_nav.columns.tolist()
        _res = []
        for fund_id in fund_nav_cols:
            benchmark_id = fund_id
            if not benchmark_id in benchmark_cols:    
                wind_class_2 = wind_class_2_dict[fund_id]
                if wind_class_2 not in self.WIND_TYPE_DICT:
                    continue
                select_type = self.WIND_TYPE_DICT[wind_class_2]
                index_price_i = index_price[self.REPLACE_DICT[select_type]].rename(benchmark_id)
                benchmark_ret = np.log(index_price_i).diff(1)
                _res.append(benchmark_ret)
        self.factor = self.factor.join(pd.concat(_res,axis=1))
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())
        cols = self.factor.columns.tolist()
        cols = [i for i in cols if isinstance(i, str)]
        self.factor = self.factor[cols]

        # fund nav is null, set benchmark null
        common_columns = list(fund_nav.columns.intersection(self.factor.columns))
        con_df = fund_nav[common_columns].isnull()
        _factor = self.factor[common_columns].copy()
        _factor[con_df] = np.nan
        _cols = self.factor.columns
        _cols_not_in = [ i for i in _cols if i not in common_columns]
        self.factor = _factor.join(self.factor[_cols_not_in])

class FundBenchmarkPrice(Factor): # based on FundBenchmarkRet

    def __init__(self):
        super().__init__(f_name='FundBenchmarkPrice', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = FundBenchmarkRet().get()
        self.factor = np.exp(self.factor.cumsum())
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())

class FundPersonalHold(Factor):

    def __init__(self):
        super().__init__(f_name='FundPersonalHold', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self.factor = DataManager.basic_data(func_name='get_history_fund_size', start_date=START_DATE).drop(columns=['_update_time'], errors='ignore')
        self.factor = self.factor.pivot(index='datetime',columns='fund_id',values='personal_holds')
        self.factor = self.data_reindex_daily_trading_day(self.factor, TradingDay().get())


'''
import surfing.data.fund_fac.basic_factors_daily as basic

attr_list = basic.__dir__()
start_idx = attr_list.index('UpdateBasicFactorStart')
update_factor_list = attr_list[start_idx+1:]

for fac in update_factor_list:
    self = eval(fac)()
    self.calc()
    self.save()
'''